<?php 
// *****************************************************************************
// Copyright 2003-2004 by A J Marston <http://www.tonymarston.net>
// Distributed under the GNU General Public Licence
// *****************************************************************************

class Default_Table {

    // member variables
    var $audit_logging;         // yes/no switch
    var $checkPrimaryKey;       // yes/no switch
    var $dbname;                // database name
    var $errors;                // array of errors
    var $fieldarray;            // array of row/field data
    var $fieldspec;             // field specifications (see class constructor)
    var $lastpage;              // last available page number in current query
    var $lookup_data;           // array of lookup data (for dropdowns, radio groups)
    var $message;               // array of messages
    var $numrows;               // number of rows retrieved
    var $pageno;                // requested page number
    var $relationship;          // relationship specifications (optional)
    var $rows_per_page;         // page size for multi-row forms
    var $skip_validation;       // yes/no switch
    var $tablename;             // table name (internal)
    var $unique_keys;           // unique key specifications (optional)
    var $updatearray;           // array that will be passed to the database for updating
    var $validate;              // validation object
    var $xref_item;             // used in getData() in many-to-many relationships
    
    // this defines the default database engine (may be changed in sub-classes)
    var $dbms_engine = 'mysql';
    
    // the following are used to construct an SQL query
    var $where;                 // passed from parent form
    var $sql_select;
    var $sql_from;
    var $sql_where;             // additional selection criteria
    var $sql_groupby;
    var $sql_having;
    var $sql_orderby;
    var $sql_orderby_seq;       // 'asc' or 'desc'
    var $default_orderby;       // default, may be overridden by $sql_orderby
    var $sql_search;            // optional search criteria from a search screen
    var $sql_search_table;      // tablename qualifier for optional search criteria
    
    // ****************************************************************************
    // class constructor
    // ****************************************************************************
    function Default_Table ()
    {
        $this->tablename       = 'default';
        $this->dbname          = 'default';
        $this->rows_per_page   = 10;
        $this->default_orderby = '';
        
        // call this method to get original field specifications
        // (note that they may be modified at runtime)
        $this->fieldspec = $this->getFieldSpec_original();
        
        // create an array of unique/candidate keys (optional)
        // each unique key may contain one or more fields
        $this->unique_keys[] = array('fieldname1', 'fieldname2');
        
        // list of optional relationships (where this table is ONE in ONE-to-MANY)
        $this->relationship[] = array('many' => 'tablename',
                                      'type' => 'nullify/delete/restricted',
                                      'fields' => array('one_id' => 'many_id'));
                                      
    } // Default_Table
    
    // ****************************************************************************
    function formatData ($fieldarray)
    // format values retrieved from the database before they are shown to the user.
    // (such as changing dates from 'CCYY-MM-DD' to 'dd Mmm CCYY'
    {
        global $dateobj;

        foreach ($fieldarray as $fieldname => $fieldvalue) {
            // only deal with fields defined in $fieldspec
            if (isset($this->fieldspec[$fieldname])) {
                // get specifications for current field
                $fieldspec = $this->fieldspec[$fieldname];
                
                if ($fieldspec['type'] == 'string') {
                    // escape any suspect characters in string fields
                    $fieldarray[$fieldname] = stripslashes($fieldvalue);
                } // if
                
                // set boolean fields to either TRUE or FALSE
                if ($fieldspec['type'] == 'boolean') {
                    if (is_True($fieldvalue)) {
                        $fieldarray[$fieldname] = $fieldspec['true'];
                    } else {
                        $fieldarray[$fieldname] = $fieldspec['false'];
                    } // if
                } // if
                
                if (isset($fieldspec['password'])) {
                    if (isset($fieldspec['hash'])) {
                        if (ereg('sha1|md5', $fieldspec['hash'])) {
                            // for this hash type do not output anything
                            $fieldarray[$fieldname] = '';
                        } // if
                    } // if
                } // if
                
                // look for fields where 'type' = 'date'
                if ($fieldspec['type'] == 'date') {
                    if (isset($fieldspec['infinityisnull']) and $fieldvalue == '9999-12-31') {
                        // this date is shown to the user as empty
                        $fieldarray[$fieldname] = '';
                    } else {
                        // convert date from internal to external format
                        $fieldarray[$fieldname] = $dateobj->getExternalDate($fieldvalue);
                    } // if
                } // if
                
                // look for fields where 'type' = 'datetime'
                if ($fieldspec['type'] == 'datetime') {
                    if (!empty($fieldvalue)) {
                        // convert date from internal to external format
                        list($date, $time) = explode(' ', $fieldvalue);
                        $date = $dateobj->getExternalDate($date);
                        $fieldarray[$fieldname] = $date . ' ' . $time;
                    } // if
                } // if
            } // if
        } // foreach 
        
        // perform any custom formatting
        $fieldarray = $this->_cm_formatData($fieldarray);
        
        return $fieldarray;
        
    } // formatData
    
    // ****************************************************************************
    function getData ($where) 
    // get data from this table using optional 'where' criteria.
    // this is formatted before being displayed to the user.
    {
        //DebugBreak();
        $this->fieldarray = array();
        $this->errors = array();
        
        // perform any custom pre-retrieve processing
        $where = $this->_cm_pre_getData($where);
        if ($this->errors) return;
        
        if ($this->checkPrimaryKey) {
            $this->errors = isPkeycomplete($where, getPkeyNames($this->fieldspec));
            if ($this->errors) return;
        } // if
        
        // get the data from the database
        $data_unformatted = $this->_dml_getData($where);

        // perform any custom post-retrieve processing
        $data_unformatted = $this->_cm_post_getData($data_unformatted, $where);
        
        // perform any formatting on the raw data
        $this->fieldarray = array();
        foreach ($data_unformatted as $rowdata) {
            $this->fieldarray[] = $this->formatData($rowdata);
        } // foreach
        
        return $this->fieldarray;
        
    } // getData
    
    // ****************************************************************************
    function getData_unformatted ($where) 
    // get data from this table using optional 'where' criteria.
    // this is returned unformatted (as read from the database).
    {
        $data_unformatted = $this->_dml_getData($where);
        
        return $data_unformatted;
        
    } // getData_unformatted
    
    // ****************************************************************************
    function getDBname () 
    // return the database name for this table.
    {
        return strtolower($this->dbname);
        
    } // getDBname
    
    // ****************************************************************************
    function getErrors () 
    // return array of error messages
    {
        if (is_string($this->errors)) {
            // convert string into an array
            $errors[] = $this->errors;
            $this->errors = $errors;
        } // if
                
        return $this->errors;
        
    } // getErrors
    
    // ****************************************************************************
    function getExtraData ($input) 
    // get additional data for this table, such as lookup lists.
    {
        // $input may be an array or a string
        if (is_array($input)) {
            $fieldarray = $input;
            $where = array2where ($input, $this->getPkeyNames());
        } else {
            if (!empty($input)) {
                $fieldarray = where2array($input);
                $fieldarray = stripOperators($fieldarray[0]);
                $where      = array2where ($fieldarray, $this->getPkeyNames());
            } else {
                $fieldarray = array();
                $where      = null;
            } // if
        } // if
        
        global $mode;
        if ($mode == 'search') {
            // do not allow any $where criteria to be changed
            foreach ($fieldarray as $fieldname => $fieldvalue) {
                if (!empty($fieldvalue)) {
                    $this->fieldspec[$fieldname]['noedit'] = 'y';
                } // if
            } // foreach
        } // if
        
        // perform custom processing (optional)
        $fieldarray = $this->_cm_getExtraData($where, $fieldarray);
        
        // store updated $fieldarray within this object
        $this->fieldarray = $fieldarray;
        
        return $fieldarray;
        
    } // getExtraData
    
    // ****************************************************************************
    function getFieldArray () 
    // return array of data that currently resides within this object
    // (usually stuff which was retrieved from the database).
    {
        if (is_long(key($this->fieldarray))) {
            // array is indexed by rownum, so return it 'as is'
            return $this->fieldarray;
        } else {
            // return unindexed array as row zero
            $array[0] = $this->fieldarray;
            return $array;
        } // if
        
    } // getFieldArray
    
    // ****************************************************************************
    function getFieldSpec () 
    // return array of current field specifications.
    {
        return $this->fieldspec;
        
    } // getFieldSpec
    
    // ****************************************************************************
    function getFieldSpec_original () 
    // set and return the original specifications for this database table
    // (before being possibly amended at runtime)
    {
        
    } // getFieldSpec_original
    
    // ****************************************************************************
    function getLastPage () 
    // return the last page number for retrieved rows.
    {
        return (int)$this->lastpage;
        
    } // getLastPage
    
    // ****************************************************************************
    function getNavButtons ($task_id) 
    // retrieve any navigation buttons for the current script.
    {
        // DebugBreak();
        $this->errors = array();
        
        $buttons = $GLOBALS['nav_buttons'];
        
        $i = 0;
        foreach ($buttons as $button_text => $script_id) {
            $array[$i]['task_id']     = $script_id;
            $array[$i]['script_id']   = $script_id;
            $array[$i]['button_text'] = $button_text;
            $i++;
        } // foreach
        
        return $array;
        
    } // getNavButtons
    
    // ****************************************************************************
    function getNumRows () 
    // return the number of rows retrived for the current page.
    {
        return (int)$this->numrows;
        
    } // getNumRows
    
    // ****************************************************************************
    function getPageNo () 
    // get current page number to be retrieved for a multi-page display.
    {
        if (empty($this->pageno)) {
            return 0;
        } else {
            return (int)$this->pageno;
        } // if
        
    } // getPageNo
    
    // ****************************************************************************
    function getPkeyNames () 
    // return the list of primary key fields in this table.
    {
        $array = getPkeyNames($this->fieldspec);
        
        return $array;
        
    } // getPkeyNames
    
    // ****************************************************************************
    function getPkeyArray () 
    // return the list of primary key values for the last selection of data
    // which was retrieved from this table.
    {
        $array = buildPkeyArray($this->fieldarray, $this->getPkeyNames());
        
        return $array;
        
    } // getPkeyNames
    
    // ****************************************************************************
    function getTableName () 
    // return the name of this table.
    {
        return strtolower($this->tablename);
        
    } // getTableName
    
    // ****************************************************************************
    function getWhere () 
    // return current selection criteria (may have been amended).
    {
        return $this->where;
        
    } // getWhere
    
    // ****************************************************************************
    function setFieldArray ($fieldarray) 
    // this allows the current data array to be set or updated.
    {
        if (empty($fieldarray)) {
            $this->fieldarray = array();
        } else {
            if (is_long(key($fieldarray))) {
                // input is indexed by row, so use it 'as is'
                $this->fieldarray = $fieldarray;
            } else {
                // input is not indexed by row, so make it row zero
                $this->fieldarray = array();
                $this->fieldarray[0] = $fieldarray;
            } // if
        } // if
        
        $this->numrows = count($this->fieldarray);
        
    } // setFieldArray
    
    // ****************************************************************************
    function setPageNo ($pageno = '1') 
    // this allows a particular page number to be selected (see getData)
    {
        $this->pageno = abs((int)$pageno);
        
    } // setPageNo
    
    // ****************************************************************************
    function setRowsPerPage ($rows_per_page) 
    // this allows the default value to be changed
    {
        $this->rows_per_page = abs((int)$rows_per_page);
        
    } // setRowsPerPage
    
    // ****************************************************************************
    function updateRecord ($fieldarray) 
    // update a record using the contents of $fieldarray.
    {
        // DebugBreak();
        if (is_long(key($fieldarray))) {
            // input is indexed by row, so extract data for 1st row only
            $array = $fieldarray[0];
            $fieldarray = $array;
        } // if
        
        $this->errors = array();
        
        // perform any custom pre-update processing
        $fieldarray = $this->_cm_pre_updateRecord($fieldarray);
        
        if (empty($this->errors)) {
            // perform standard declarative checks on input data
            $updatearray = $this->validateUpdateArray($fieldarray);
        } // if
        
        if (empty($this->errors)) {
            // build 'where' string using values for primary key
            $where = array2where($updatearray, $this->getPkeyNames()); 
            // obtain copy of original record from database
            $originaldata = $this->_dml_ReadBeforeUpdate($where);
            if ($this->numrows <> 1) {
                $this->errors[] = "Could not locate original $this->tablename record for updating";
            } else {
                // use only 1st row in $originaldata
                $originaldata = $originaldata[0]; 
                // insert values missing from update array before validation
                $updatearray = array_merge($originaldata, $updatearray);
            } // if
        } // if
        
        if (empty($this->errors)) {
            if ($this->skip_validation) {
                // do not perform any custom validation
            } else {
                // perform any custom validation/processing before update
                $updatearray = $this->_cm_commonValidation($updatearray, $originaldata);
                if (empty($this->errors)) {
                    $updatearray = $this->_cm_validateUpdate($updatearray, $originaldata);
                } // if
            } // if
        } // if
        
        // everything OK so far, so update the database
        if (empty($this->errors)) {
            // pass both the updated and the original data for processing
            $this->_dml_updateRecord($updatearray, $originaldata);
        } // if
        
        // replace any fields excluded from $updatearray
        $fieldarray = array_merge($fieldarray, $updatearray);
        
        if (empty($this->errors)) {
            if ($this->skip_validation) {
                // do not perform any custom processing
            } else {
                // perform any custom post-update processing
                $fieldarray = $this->_cm_post_updateRecord($fieldarray, $originaldata);
            } // if
        } // if
        
        // turn this flag off
        $this->skip_validation = false;
        
        // format data for output to the user
        $fieldarray = $this->formatData($fieldarray);
        
        // store updated $fieldarray within this object
        $this->fieldarray = $fieldarray;
        
        return $fieldarray;
        
    } // updateRecord
    
    // ****************************************************************************
    function validateUpdateArray ($fieldarray) 
    // validate contents of $fieldarray prior to an UPDATE
    {
        //DebugBreak();
        if (!isset($this->validate) OR !is_object($this->validate)) {
            require_once 'std.validation.class.inc';
            $this->validate = new Validation;
        } // if

        $array = $this->validate->validateUpdateArray($fieldarray, $this->fieldspec, $this);

        $this->errors = $this->validate->getErrors();

        return $array;
        
    } // validateUpdateArray
    
    // ****************************************************************************
    // methods beginning with '_cm_' are designed to be customised as required
    // ****************************************************************************
    function _cm_commonValidation ($fieldarray, $originaldata) 
    // perform validation that is common to INSERT and UPDATE.
    {
        return $fieldarray;
        
    } // _cm_commonValidation
    
    // ****************************************************************************
    function _cm_formatData ($fieldarray) 
    // perform custom formatting before values are shown to the user.
    {
        return $fieldarray;
        
    } // _cm_formatData
    
    // ****************************************************************************
    function _cm_getExtraData($where, $fieldarray) 
    // Perform custom processing for the getExtraData method.
    // $where = a string in SQL 'where' format.
    // $fieldarray = the contents of $where as an array.
    {
        return $fieldarray;
        
    } // _cm_getExtraData
    
    // ****************************************************************************
    function _cm_post_getData ($rowdata, &$where) 
    // perform custom processing after database record(s) are retrieved.
    // NOTE: $where is passed BY REFERENCE so that it may be modified.
    {
        return $rowdata;
        
    } // _cm_post_getData
    
    // ****************************************************************************
    function _cm_post_updateRecord ($fieldarray, $old_data) 
    // perform custom processing after database record is updated.
    {
        return $fieldarray;
        
    } // _cm_post_updateRecord
    
    // ****************************************************************************
    function _cm_pre_getData ($where) 
    // perform custom processing before database record(s) are retrieved.
    {
        return $where;
        
    } // _cm_pre_getData
    
    // ****************************************************************************
    function _cm_pre_updateRecord ($fieldarray) 
    // perform custom processing before database record is updated.
    // errors are added to $this->errors.
    {
        return $fieldarray;
        
    } // _cm_pre_updateRecord
    
    // ****************************************************************************
    function _cm_validateUpdate ($fieldarray, $originaldata) 
    // perform custom validation before update.
    {
        return $fieldarray;
        
    } // _cm_validateUpdate
    
    // ****************************************************************************
    // methods beginning with '_dml_' are for calling the DML object
    // ****************************************************************************
    function _dml_getCount ($where)
    // Get count of recors which match criteria in $where.
    {
        // DebugBreak();
        $DML = &$this->_getDBMSengine();
        
        $count = $DML->getCount($this->dbname, $this->tablename, $where);
        
        $this->errors = array_merge($DML->getErrors(), $this->errors);
        
        return $count;
        
    } // _dml_getCount
    
    // ****************************************************************************
    function _dml_getData ($where)
    // Get data from the specified database table.
    // Results may be affected by $where and $pageno.
    {
        // DebugBreak();
        $this->errors = array();
        
        $DML = &$this->_getDBMSengine();
        
        $DML->fieldspec     = $this->fieldspec;
        $DML->pageno        = $this->pageno;
        $DML->rows_per_page = $this->rows_per_page;
        $DML->sql_from      = $this->sql_from;
        $DML->sql_groupby   = $this->sql_groupby;
        $DML->sql_having    = $this->sql_having;
        
        if (isset($this->sql_orderby)) {
            $DML->sql_orderby = $this->sql_orderby;
        } else {
            $DML->sql_orderby = $this->default_orderby;
        } // if
        
        $DML->sql_orderby_seq = $this->sql_orderby_seq;
        $DML->sql_select      = $this->sql_select;
        $DML->sql_where       = $this->sql_where;
        
        $array = $DML->getData($this->dbname, $this->tablename, $where);
        
        $this->errors   = array_merge($DML->getErrors(), $this->errors);
        $this->numrows  = $DML->getNumRows();
        $this->pageno   = $DML->getPageNo();
        $this->lastpage = $DML->getLastPage();
        
        return $array;
        
    } // _dml_getData
    
    // ****************************************************************************
    function _dml_ReadBeforeUpdate ($where)
    // Read a single record just before it is updated.
    // The primary key should be supplied in $where.
    {
        // DebugBreak();
        $DML = &$this->_getDBMSengine();
        
        $DML->fieldspec       = $this->fieldspec;
        $DML->pageno          = 1;
        $DML->rows_per_page   = 10;
        $DML->sql_from        = null;
        $DML->sql_groupby     = null;
        $DML->sql_having      = null;
        $DML->sql_orderby     = null;
        $DML->sql_orderby_seq = null;
        $DML->sql_select      = null;
        $DML->sql_where       = null;
        
        $array = $DML->getData($this->dbname, $this->tablename, $where);
        
        $this->errors = array_merge($DML->getErrors(), $this->errors);
        
        $this->numrows = $DML->getNumRows();
        
        return $array;
        
    } // _dml_ReadBeforeUpdate
    
    // ****************************************************************************
    function _dml_updateRecord ($fieldarray, $oldarray)
    // update the record contained in $fieldarray.
    {
        // DebugBreak();
        $DML = &$this->_getDBMSengine();
        
        // use ORIGINAL, not CURRENT specifications for this database table
        $DML->fieldspec = $this->getFieldSpec_original();
        
        $DML->unique_keys = $this->unique_keys;
        
        $DML->updateRecord($this->dbname, $this->tablename, $fieldarray, $oldarray);
        
        $this->errors = array_merge($DML->getErrors(), $this->errors);
        
        $this->numrows = $DML->getNumRows();
        
        return;
        
    } // _dml_updateRecord
    
    // ****************************************************************************
    function &_getDBMSengine ()
    // get/create the object that deals with this database engine.
    {
        // DebugBreak();
        global $dbms_engine;
        
        foreach ($dbms_engine as $name => $object) {
            if ($name == $this->dbms_engine) {
                $DML = &$object;
                return $DML;
            } // if
        } // foreach
        
        // create a new object for this database engine
        require_once "dml.$this->dbms_engine.class.inc";
        $dbms_engine[$this->dbms_engine] = &new $this->dbms_engine;
        
        return $dbms_engine[$this->dbms_engine];
        
    } // _getDBMSengine
    
    // ****************************************************************************
    function __sleep () 
    // perform object clean-up before serialization.
    {
        // get associative array of class variables
        $object_vars = get_object_vars($this);
        
        // remove/clear unwanted variables
        unset($object_vars['errors']);
        
        // convert to indexed array
        $object_vars = array_keys($object_vars);
        
        return $object_vars;
        
    } // __sleep
    
// ****************************************************************************
} // end class
// ****************************************************************************

?>
